@@ -1,5 +1,14 @@
 #!/usr/bin/perl
 
+# Main qemu-web-desktop service script.
+#
+#   You must edit the /etc/qemu-web-desktop/config.pl file to configure the service.
+#   Think about editing the /usr/share/qemu-web-desktop/html/desktop/index.html
+#   to comment/un-comment sections (GPU, oneshot, user scripts, ...).
+#
+#   When relevant, the equivalent OpenStack naming is indicated (as documented 
+#   in https://docs.openstack.org/nova/latest/configuration/config.html)
+#
 # This script is triggered by a FORM or runs as a script.
 # to test this script, launch from the project root level something like:
 #
@@ -40,7 +49,7 @@
 #   libcgi-pm-perl liblist-moreutils-perl libnet-dns-perl libjson-perl \
 #   libproc-background-perl libproc-processtable-perl libemail-valid-perl \
 #   libnet-smtps-perl libmail-imapclient-perl libnet-ldap-perl libemail-valid-perl \
-#   libwww-perl
+#   libwww-perl liburi-perl
 #
 # (c) 2020 Emmanuel Farhi - GRADES - Synchrotron Soleil. AGPL3.
 # https://gitlab.com/soleil-data-treatment/remote-desktop
@@ -80,6 +89,9 @@ use Mail::IMAPClient;   # libmail-imapclient-perl   for imap user check
 use Net::LDAP;          # libnet-ldap-perl          for ldap user check
 use Email::Valid;       # libemail-valid-perl
 use LWP::UserAgent;     # libwww-perl               for new and get(URL)
+use Net::Ping;          # perl                      to test server up-down
+use URI;                # liburi-perl               to split URL into parts
+
 
 # see http://honglus.blogspot.com/2010/08/resolving-perl-cgi-buffering-issue.html
 $| = 1;
@@ -109,6 +121,20 @@ $config{version}                  = "22.04.22";  # year.month.day
 
 require '/etc/qemu-web-desktop/config.pl';
 
+# search detached GPU (via vfio-pci). Only use video part, no audio.
+{
+  # can be a VGA adaptor or a 'controller'
+  my ($device_pci1, $device_model1, $device_name1) = pci_devices("lspci -nnk","vga","vfio");
+  my ($device_pci2, $device_model2, $device_name2) = pci_devices("lspci -nnk","controller","vfio");
+  
+  my @gpu_pci  = (@$device_pci1,   @$device_pci2);
+  my @gpu_model= (@$device_model1, @$device_model2);
+  my @gpu_name = (@$device_name1,  @$device_name2);
+  $config{gpu_model}             = \@gpu_model;
+  $config{gpu_name}              = \@gpu_name;
+  $config{gpu_pci}               = \@gpu_pci;
+}
+
 # ------------------------------------------------------------------------------
 # update config with input arguments from the command line (when run as script)
 # ------------------------------------------------------------------------------
@@ -116,12 +142,9 @@ require '/etc/qemu-web-desktop/config.pl';
 # the 'service_monitor' can be set to true to generate a list of running sessions.
 # each of these can display its configuration, and be stopped/cleaned.
 $config{service_monitor}          = 0;
-
-$config{session_stop}             = ""; # send a json ref, stop service
-
+$config{session_stop}             = ""; # contains a json ref, stop service
 $config{session_purge}            = 0;  # when true stop/clean all
-
-$config{oneshot} = 0;
+$config{oneshot}                  = 0;
 
 for(my $i = 0; $i < @ARGV; $i++) {
   $_ = $ARGV[$i];
@@ -182,6 +205,7 @@ my %session;
 
 # transfer defaults
 $session{machine}     = $config{machine};
+# OpenStack DEFAULT.instance_name_template ?
 $session{dir_snapshot}= tempdir(TEMPLATE => "$config{service}" . "_XXXXXXXX", 
   DIR => $config{dir_snapshots}) || die "Can not create snapshot directory in $config{dir_snapshots}.";
 $session{name}        = File::Basename::fileparse($session{dir_snapshot});
@@ -201,6 +225,8 @@ $session{config_script}= ""; # a file/URL to execute before starting the VM
 $session{snapshot_lifetime} = $config{snapshot_lifetime}; # default [sec]
 $session{oneshot}     = $config{oneshot};
 
+$session{user_email}  = "";
+
 $session{date}        = localtime();
 # see https://www.oreilly.com/library/view/perl-cookbook/1565922433/ch11s03.html#:~:text=To%20append%20a%20new%20value,values%20for%20the%20same%20key.
 #   on how to handle arrays in a hash.
@@ -210,7 +236,7 @@ $session{pid}         = ();     # we search all children in session_stop
 # push @{ $session{pid} }, $$;    # do NOT add our own PID = common Perl for all cgi's
 $session{pid_wait}    = $$;     # PID to wait for (daemon).
 $session{port}        = $config{service_port}; # e.g. 6080 for websockify
-$session{qemuvnc_ip}  = "127.0.0.1";
+$session{qemuvnc_ip}  = "127.0.0.1"; # OpenStack vnc.server_listen vnc.proxyclient_address
 # cast a random token key for VNC: 8 random chars in [a-z A-Z digits]
 sub rndStr{ join'', @_[ map{ rand @_ } 1 .. shift ] };
 $session{token_vnc}   = rndStr (8, 'a'..'z', 'A'..'Z', 0..9);
@@ -234,7 +260,7 @@ $session{remote_host} = $q->remote_host(); # the 'client'
 if ($session{remote_host} =~ "::1") {
   $session{remote_host} = "localhost";
 }
-$session{server_name} = $q->server_name(); # the 'server'
+$session{server_name} = $q->server_name(); # the 'server' OpenStack DEFAULT.host
 if ($session{server_name} =~ "::1") {
   $session{server_name} = "localhost";
 }
@@ -253,16 +279,24 @@ for ('machine','user','cpu','memory','gpu','snapshot_lifetime',
   }
 }
 
-# these are the "input" to collect from the HTML FORM.
-# count how many parameters are undef. All will when running as script.
+# these are the "input" to collect from the HTML FORM. Store in $session
 for ('machine','user','cpu','memory','gpu','config_script','snapshot_lifetime',
-  'session_stop','session_purge','password','oneshot') {
+  'session_stop','session_purge','password','oneshot','fallback_servers') {
   my $val = $q->param($_);
   if (defined($val)) {
     $session{$_} = $val;
   } 
 }
 
+# we get the list of fallback servers to be used when current machine is over-subscribed.
+# The POST message will be updated with a reduced fallback_servers list
+# and sent via LWP::UserAgent -> post
+if (not defined $session{fallback_servers} and defined $config{fallback_servers}) {
+  $session{fallback_servers} = $config{fallback_servers};
+}
+my $need_fallback_server = 0;
+
+
 # check if the 'Manage button' was hit, instead of Create
 if (defined $q->param('manage')) {
   $session{machine} = 'monitor';
@@ -326,38 +360,45 @@ if ($session{machine} =~ 'purge' or $session{session_purge}) {
 # handle 'admin'actions
 if (not $error and ($config{service_monitor} or $config{session_purge})) {
   (my $out, my $err) = session_authenticate(\%config, \%session);
-  my $user  = $session{user};
-  if ($session{session_purge}) { $user = $session{session_purge}; }
-  my @admin = @{ $config{user_admin} };
-  my $isadmin = 0;
-  $session{machine} = "administration";
-  if ($session{runs_as_cgi} and not $err and not grep( /^$user$/, @admin ) ) {
-    $isadmin = 0;
-  } else { $isadmin = 1; }
   
-  if ($config{service_monitor}) {
-    $output = service_monitor(\%config, $output, $isadmin, $session{user});
-  } elsif ($config{session_purge}) {
-    if (not $isadmin) {
-      $err .= "User $user is not among the 'user_admin' list";
-      $output .= "</ul><h1>[ERROR] $err</h1></body></html>";
-    } else {
-      $config{snapshot_lifetime} = 1;
-      print STDERR $0." [$session{date}] Stopping all session, requested by $user\n";
-      service_housekeeping(\%config);
-      $output .= '</ul><h1>OK: Purged all</h1>';
-    }
-  }
-  # display...
-  print "Content-type:text/html\r\n\r\n";
-  print "$output\n\n";
-  if ($r->can("rflush")) { 
-    $r->rflush; 
-  }
+  if (not $err) { # must be identified
+    my $user  = $session{user};
+    if ($session{session_purge}) { $user = $session{session_purge}; }
+    my @admin = @{ $config{user_admin} };
+    my $isadmin = 0;
+    $session{machine} = "administration";
+    if ($session{runs_as_cgi} and not grep( /^$user$/, @admin ) ) {
+      $isadmin = 0;
+    } else { $isadmin = 1; }
     
-  exit;
+    if ($config{service_monitor}) {
+      $output = service_monitor(\%config, $output, $isadmin, $user);
+    } elsif ($config{session_purge}) {
+      if (not $isadmin) {
+        $err .= "User $user is not among the 'user_admin' list";
+      } else {
+        $config{snapshot_lifetime} = 1;
+        print STDERR $0." [$session{date}] Stopping all session, requested by $user\n";
+        service_housekeeping(\%config);
+        $output .= '</ul><h1>OK: Purged all</h1>';
+      }
+    }
+  } 
+  
+  if (not $err) {
+    # display...
+    print "Content-type:text/html\r\n\r\n";
+    print "$output\n\n";
+    if (defined($r) and $r->can("rflush")) { 
+      $r->rflush; 
+    }
+    exit;
+  } else {
+    $error .= $err;
+  }
 }
 
+
 # ------------------------------------------------------------------------------
 # Session checks: cpu, memory, disk, VM, config_script
 # ------------------------------------------------------------------------------
@@ -368,32 +409,135 @@ if ($config{service_monitor} or $config{session_purge}) {
 } else {
   if ($config{session_nb} > $config{service_max_session_nb}) {
     $error .= "Too many active sessions $config{session_nb}. Max $config{service_max_session_nb}. Try again later.";
+    $need_fallback_server = 1;
   }
   if (Sys::CPU::cpu_count()-Sys::CpuLoad::load() < $session{cpu}) {
     $error .= "Not enough free CPU's. Try again later.\n";
+    $need_fallback_server = 1;
   }
   if (Sys::CpuLoad::load() / Sys::CPU::cpu_count() > $config{service_max_load}) {
     $error .= "Server load exceeded. Try again later.\n";
+    $need_fallback_server = 1;
   }
   my ($total_memory_kb, $free_memory_kb, $avail_memory_kb) = get_freemem();
   if ($avail_memory_kb / 1024 < $session{memory}) { # in MB
     $error .= "Not enough free memory. Try again later.\n";
+    $need_fallback_server = 1;
   }
   if (not -e "$config{dir_machines}/$session{machine}") {
     $error .= "Can not find virtual machine.\n";
   }
-  if ($session{config_script}) {
-    if (length($session{config_script}) > 256) {
-      $error .= "Configuration script specification is too long (>256).\n";
-    }
-    if($session{config_script} !~ /^(?:[a-z0-9\-\._~!\$&'\(\)\*+,;=:\/?@]|%[0-9A-F]{2})*$/i) {
-      $error .= "Invalid configuration script specification (file/URL).\n";
-    }
+  (my $nb_session, my $nb_cpu, my $nb_mem) = service_check_user_max(\%config, $session{user});
+  if ($nb_session > $config{service_max_session_nb_per_user}) {
+    $error .= "Too many opened sessions ($nb_session) for user. Try again later.\n";
+    $need_fallback_server = 1;
+  }
+  if ($nb_cpu/Sys::CPU::cpu_count() > $config{service_max_cpu_fraction_nb_per_user}) {
+    $error .= "Too many used CPUs ($nb_cpu) for user. Try again later.\n";
+    $need_fallback_server = 1;
+  }
+  if ($nb_mem/($avail_memory_kb / 1024) > $config{service_max_mem_fraction_nb_per_user}) {
+    $error .= "Too much memory ($nb_mem) used for user. Try again later.\n";
+    $need_fallback_server = 1;
   }
 }
 
+# find a free GPU when requested
+if (defined($session{gpu}) and $session{gpu} =~ /yes|gpu|true|1/i) { 
+  # look for detached GPU PCI, and check if it is used by a running session
+  $session{gpu} = "";
+  foreach my $pci (@{ $config{gpu_pci} }) {
+    if (not $session{gpu} and not session_use_gpu(\%config, $pci)) {
+      $session{gpu} = $pci; # this is what we need to pass to qemu
+    }
+  }
+  if ($session{gpu}) {
+    $output .= "<li>$ok Assigned GPU at PCI $session{gpu}.</li>\n";
+  } else {
+    $error .= "Can not find a free GPU as requested. Try again without.\n";
+    $need_fallback_server = 1;
+  }
+}
+
+# check if the current server is oversubscribed and there are fallback servers
+if ($need_fallback_server and length($session{fallback_servers})) {
+
+  # for loop: search a valid server separated with ','
+  my $host="";
+  my $fallback_server = "";
+
+  for $host (split /,/, $session{fallback_servers}) {
+    # get server to try: test if $host is not empty and active
+    my $p   = Net::Ping->new();
+    my $url = URI->new($host);
+    my $h   = $host;
+    if ($url->can("host")) { $h = $url->host; }
+    if (length($host) and $url and $p->ping($h, 0.1)) {
+      # server is active, remove it from the list of fallback_servers (replace with ' ')
+      $session{fallback_servers} =~ s/$host/ /ig;
+      if (not $url->scheme) {
+        # add protocol/scheme
+        $host = "$config{http}://$host";
+      }
+      if (not $url->path and IS_CGI) {
+        # add CGI path
+        $host .= $ENV{SCRIPT_NAME};
+      }
+      $fallback_server = $host;
+      last;    
+    }
+    $p->close();
+  }
+  # end for. $host now contains the server to forward the POST to.
+  
+  if ($fallback_server) {
+    # update the fallback_servers in the POST message
+    print STDERR "$0: forwarding request to: $fallback_server\n";
+    
+    $q->param( fallback_servers => $session{fallback_servers} );
+    
+    # send POST message to $host server
+    # ignore SSL certificate failures (we trust our fallback servers)
+    my $ua = LWP::UserAgent->new(
+      ssl_opts => { SSL_verify_mode => 0, verify_hostname => 0}); 
+    # use a specific proxy, or system configuration (e.g. HTTP_PROXY env var)
+    if ($config{service_proxy}) {
+      $ua->proxy(['http', 'ftp', 'https'], $config{service_proxy});
+    } else { $ua->env_proxy; }
+    
+    # forward POST with updated fallback_servers list
+    my %params; # hash to store CGI arguments from the form
+    foreach my $p ($q->param) {
+      $params{$p} = $q->param($p);
+    }
+    
+    # send to fallback server
+    my $response = $ua->post($fallback_server, \%params); 
+
+    # display message
+    if ($response->is_success) {
+      my $output = $response->as_string();
+      # must clean the message up to '<!DOCTYPE html'
+      my $doctype_index = index($output, '<!DOCTYPE html'); # start at 0
+      if ($doctype_index > -1) { 
+        $output = substr($output, $doctype_index); # skip protocol info
+      }
+      print "Content-type:text/html\r\n\r\n";
+      print "$output\n\n";
+      $session{server_name} = $fallback_server;
+      session_save(\%session);
+
+      # exit
+      exit
+    } else {
+      print STDERR "$0: forward not succesful from $fallback_server:\n";
+      print STDERR $response->as_string();
+    }
+  } # else proceed with error message: no localhost cpu/mem/gpu resource left
+}
+
 # ------------------------------------------------------------------------------
-# User credentials checks
+# User credentials checks, find communication ports
 # ------------------------------------------------------------------------------
 {
   (my $out, my $err) = session_authenticate(\%config, \%session);
@@ -449,22 +593,6 @@ if (not defined($port_vnc)) {
 }
 $session{port_vnc} = $port_vnc;
 
-# find a free GPU when requested
-if (defined($session{gpu}) and $session{gpu} =~ /yes|gpu|true|1/i) { 
-  # look for detached GPU PCI, and check if it is used by a running session
-  $session{gpu} = "";
-  foreach my $pci (@{ $config{gpu_pci} }) {
-    if (not $session{gpu} and not session_use_gpu(\%config, $pci)) {
-      $session{gpu} = $pci; # this is what we need to pass to qemu
-    }
-  }
-  if ($session{gpu}) {
-    $output .= "<li>$ok Assigned GPU at PCI $session{gpu}.</li>\n";
-  } else {
-    $error .= "Can not find a free GPU as requested. Try again without.\n";
-  }
-}
-
 # ==============================================================================
 # DO the work
 # ==============================================================================
@@ -498,50 +626,14 @@ if (not $error) {
   }
 } # Create snapshot
 
-# LAUNCH optional 'configuration' script
+# LAUNCH optional 'configuration' scripts (system wide, and user level)
+if (not $error and $config{config_script}) {
+  foreach my $c (@{ $config{config_script} }) {
+    session_run_script(\%config, \%session, $c);
+  }
+}
 if (not $error and $session{config_script}) {
-  # add 'file://' for local file 
-  if (-e "$session{config_script}") {
-    $session{config_script} = "file://".$session{config_script};
-  }
-  # get script from $session{config_script}
-  my $content = undef;
-  # we try a few times to get the content. A single attempt is often not enough.
-  for (1..5) {
-    my $ua = LWP::UserAgent->new(timeout => 2);
-    # use a specific proxy, or system configuration (e.g. HTTP_PROXY env var)
-    if ($config{service_proxy}) {
-      $ua->proxy(['http', 'ftp', 'https'], $config{service_proxy});
-    } else { $ua->env_proxy; }
-    my $response = $ua->get($session{config_script});
-    if (not $response->is_error) {
-      $content = $response->content;
-      last;
-    }
-  }
-    
-  if (not defined ($content) or not $content) {
-      $output .= "<font color=orange>[WARNING]</font> Could not get start script <a href='$session{config_script}' target=_blank>$session{config_script}</a>\n";
-      print STDERR $0.": $session{name}: $session{date}: $session{user} can not get config_script=$session{config_script}\n";
-  } else {
-    # remove invalid characters
-    $content=~s/[^\x00-\x7f]//g;
-    if (length($content) > 65536) {
-      $error .= "Configuration script content is too long (>65536).";
-    } else {
-      # store script in our 'snapshot' directory
-      open my $fh, ">", "$session{dir_snapshot}/configuration_script";
-      print $fh "$content\n";
-      close $fh;
-      # output script into apache2/log
-      print STDERR $0.": $session{name}: $session{date}: $session{user} uses config_script=$session{config_script} -> $session{dir_snapshot}/configuration_script\n";
-      print STDERR "$content\n";
-      # call virt-customize: start VM, execute script, stop it. Only runs on stopped VMs.
-      # virt-customize -a vm.qcow2 --run snapshot/script
-      my $cmd = "virt-customize -a $session{snapshot} --firstboot $session{dir_snapshot}/configuration_script";
-      my $res = `$cmd`;
-    }
-  }
+  session_run_script(\%config, \%session, $session{config_script});
 }
 
 # LAUNCH CLONED VM -------------------------------------------------------------
@@ -717,12 +809,12 @@ if (not $error) {
   }
 }
 
-# save session info
-session_save(\%session);
-
 # COMPLETE OUTPUT MESSAGE ------------------------------------------------------
 if (not $error) {
 
+  # save session info only when no error if found
+  session_save(\%session);
+
   # $output .= "<li>$ok No error, all is fine.</li>\n";
   $output .= "<li>$ok Connect to your machine at <a href=$session{url} target=_blank>$session{url}</a>.</li>\n";
   if ($session{token_vnc} and $config{service_port_multiple}) {
@@ -773,7 +865,10 @@ if (not $error) {
     <li>The virtual machine is created on request, and not kept. 
       Your work <b>must be saved elsewhere</b> 
       (e.g. mounted disk, ssh/sftp, Dropbox, OwnCloud...).</li>
-    <li>To <b>kill</b> this session, click on <form action=\"/cgi-bin/qemu-web-desktop.pl\" method=post target=_blank><input type=hidden name=session_stop value=\"$session{name}\"><input type=submit value=STOP style=\"background-color:red\"></input></form></li>
+    <li>To <b>kill</b> this session, click on <form action="$config{http}://$session{server_name}/cgi-bin/qemu-web-desktop.pl" method=post target=_blank>
+    <input type=hidden name=session_stop value="$session{name}">
+    <input type=image alt="stop" src="$config{http}://$session{server_name}/qemu-web-desktop/images/stop-button.png" title=STOP height="32">
+    </form></li>
     </ul></p>
 
     <hr>
@@ -801,24 +896,23 @@ END_HTML
 }
 
 # send message via email when possible
-if ($config{check_user_with_email}) {
+if ($config{check_user_with_email} and not $error) {
   # send the full output message (with URL)
   session_email(\%config, \%session, $output);
 }
 
 # write an HTML file to store the message output
-{
+# display the output message (redirect) ----------------------------------------
+if (not $error) {
   open my $fh, ">", $session{output};
   print $fh "$output\n";
   close $fh;
-}
-
-# display the output message (redirect) ----------------------------------------
-print STDERR $0.": $session{name}: $session{date}: START $session{machine} by $session{user}\n";
-print STDERR $0.": $session{name}: json:  $session{json}\n";
-print STDERR $0.": $session{name}: URL:   $session{url}\n";
-print STDERR $0.": $session{name}: PIDs:  @{$session{pid}}\n" if ($session{pid});
-if ($error) {
+  
+  print STDERR $0.": $session{name}: $session{date}: START $session{machine} by $session{user}\n";
+  print STDERR $0.": $session{name}: json:  $session{json}\n";
+  print STDERR $0.": $session{name}: URL:   $session{url}\n";
+  print STDERR $0.": $session{name}: PIDs:  @{$session{pid}}\n" if ($session{pid});
+} else {
   print STDERR $0.": $session{name}: $session{date}: ERROR $error\n";
 }
 if ($session{runs_as_cgi}) {
@@ -831,7 +925,7 @@ if ($session{runs_as_cgi}) {
   print "$output\n\n";
     
   # the 1st argument of CGI is an Apache RequestIO
-  if ($r->can("rflush")) { 
+  if (defined($r) and $r->can("rflush")) { 
     $r->rflush; 
   }
 }
@@ -857,7 +951,9 @@ if (not $error and $proc_novnc and $proc_qemu and $session{oneshot}) {
 # - session_save
 # - session_load
 # - session_stop
+# - session_run_script
 # - service_housekeeping
+# - service_check_user_max
 # - service_monitor
 # - session_email
 # - session_authenticate
@@ -947,6 +1043,91 @@ sub session_stop {
   
 } # session_stop
 
+# session_run_script(\%config, \%session, $script): install a script for boot in VM
+# The script can be specifyed as:
+#   "http://some/url"
+#   "/some/local/path/to/script"
+#   "exec: some commands" separated with EOL (\n) or ';'
+#   "virt-customize: commands" separated with EOL (\n)
+sub session_run_script {
+  my $config_ref  = shift;
+  my $session_ref = shift;
+  my $script      = shift;
+  
+  if (not $session_ref) { return ""; }
+  if (not $config_ref)  { return ""; }
+  my %config  = %{ $config_ref };
+  my %session = %{ $session_ref };
+  
+  if (length($script) > $config{service_max_script_length}) {
+    return "";
+  }
+  
+  my $content     = undef;
+  my $virt_customize_commands = 0;
+  
+  # add 'file://' for local file
+  if (rindex($script, "bash:", 0) == 0 or rindex($script, "exec:", 0) == 0) {
+    $content = substr($script, 5);
+  }
+  elsif (rindex($script, "virt-customize:", 0) == 0) {
+    $content = substr($script, 15);
+    $virt_customize_commands = 1;
+  }
+  elsif (-e "$script") {
+    $script = "file://".$script;
+  }
+  
+  # get script from $script URL
+  # we try a few times to get the content. A single attempt is often not enough.
+  if (not defined ($content) or not $content) {
+    for (1..5) {
+      my $ua = LWP::UserAgent->new(timeout => 2);
+      # use a specific proxy, or system configuration (e.g. HTTP_PROXY env var)
+      if ($config{service_proxy}) {
+        $ua->proxy(['http', 'ftp', 'https'], $config{service_proxy});
+      } else { $ua->env_proxy; }
+      my $response = $ua->get($script);
+      if (not $response->is_error) {
+        $content = $response->content;
+        last;
+      }
+    }
+  }
+    
+  if (defined ($content) and length($content)) {
+    # remove invalid characters
+    $content=~s/[^\x00-\x7f]//g;
+    if (length($content) < $config{service_max_script_length}) {
+
+      # replace @symbols@ (e.g. for auto login)
+      $content=~s/\@USER\@/$session{user}/g;
+      $content=~s/\@SESSION_NAME\@/$session{name}/g;
+      $content=~s/\@VM\@/$session{machine}/g;
+      
+      # store script in our 'snapshot' directory for local exec.
+      open my $fh, ">", "$session{dir_snapshot}/configuration_script";
+      print $fh "$content\n";
+      close $fh;
+      
+      # output script into apache2/log
+      print STDERR $0.": $session{name}: $session{date}: $session{user} uses config_script=$script -> $session{dir_snapshot}/configuration_script\n";
+      print STDERR "$content\n";
+      
+      # call virt-customize: start VM, execute script, stop it. Only runs on stopped VMs.
+      # virt-customize -a vm.qcow2 --run snapshot/script
+      my $cmd = "virt-customize -a $session{snapshot} ";
+      if ($virt_customize_commands == 0) {
+        $cmd .= "--firstboot $session{dir_snapshot}/configuration_script";
+      } else {
+        $cmd .= "--commands-from-file $session{dir_snapshot}/configuration_script";
+      }
+      my $res = `$cmd`;
+      return $res;
+    }
+  }
+} # session_run_script
+
 # service_housekeeping(\%config): scan 'snapshot' and 'cfg' directories.
 #   - kill over-time sessions
 #   - check that 'snapshots' have a 'cfg'.
@@ -986,7 +1167,7 @@ sub service_housekeeping {
             print STDERR $0." [$now] housekeeping: $cfg/$snaphot_name.json (exceeded life-time)\n";
             session_stop($session_ref);
           }
-        }
+        } else { rmtree( $snapshot ) || print STDERR "Failed removing $snapshot"; }
       }
     }
   }
@@ -1000,7 +1181,7 @@ sub service_housekeeping {
   return ($err,$nb);
 } # service_housekeeping
 
-# service_monitor(\%config, $out, $isadmin): present a list of running sessions 
+# service_monitor(\%config, $out, $isadmin, $user): present a list of running sessions 
 #   as well as the server usage and history.
 #   return appended string $out
 sub service_monitor {
@@ -1040,96 +1221,70 @@ sub service_monitor {
   my $cfg     = $config{dir_cfg};
   my $service = $config{service};
   
-  # {name} {machine} {user} {date} {cpu} {mem} {lifetime} {url} {PIDs}
-  $out .= "<br><hr><br><h1>Active sessions for $user</h1><table  border='1'>\n";
-  $out .= "<tr><th>Start Date</th><th>Name</th><th>Machine</th><th>User</th>";
-  $out .= "<th>CPU </th><th>Memory (MB)</th><th>GPU</th><th>lifetime (h)</th><th>URL</th>";
-  $out .= "<th>PID's </th></tr>\n";
+  # iter=1: normal user; 2: admin user
+  for my $iter (1..2) {
   
-  foreach my $snapshot (glob("$dir/$service"."_*")) {
-    if (-d $snapshot) { # is a snapshot directory
-      my $snaphot_name = fileparse($snapshot); # just the session name
-      if (-e "$cfg/$snaphot_name.json") {
-        my $session_ref = session_load(\%config, "$cfg/$snaphot_name.json");
-        if ($session_ref) {
-          my %session = %{ $session_ref };
-          if ($session{pid} and $user =~ $session{user}) {
-            my @pids = @{ $session{pid} };
-            $out .= "<tr>";
-            $out .= "<td>$session{date}</td>";
-            $out .= "<td>$session{name}</td>";
-            $out .= "<td>$session{machine}</td>";
-            $out .= "<td>$session{user}</td>";
-            $out .= "<td>$session{cpu}</td>";
-            $out .= "<td>$session{memory}</td>";
-            if ($session{gpu}) {
-              $out .= "<td>$session{gpu}</td>";
-            } else { $out .= "<td></td>"; }
-            my $lt = $session{snapshot_lifetime}/3600;
-            $out .= "<td>$lt</td>";
-            if ($session{oneshot}) {
-              $out .= "<td>one-shot";
-            } else {
-              $out .= "<td><a href='$session{url}' target=_blank><p style=\"color:green;\"><b>CONNECT</b></p></a>";
-              if ($session{token_vnc} && $config{service_port_multiple}) {
-                $out .= " $session{token_vnc}";
+    # {name} {machine} {user} {date} {cpu} {mem} {lifetime} {url} {PIDs}
+    if ($iter == 1) {
+      $out .= "<br><hr><br><h1>Active sessions for $user</h1><table  border='1'>\n";
+      $out .= "<tr><th>Start Date</th><th>Name</th><th>Machine</th><th>User</th>";
+      $out .= "<th>CPU </th><th>Memory (MB)</th><th>GPU</th><th>lifetime (h)</th><th>URL</th>";
+      $out .= "<th>PID's </th></tr>\n";
+    } elsif ($iter == 2 and $isadmin) {
+      $out .= "<br><hr><br><h1>Active sessions [$config{session_nb}] (administration)</h1><table  border='1'>\n";
+      $out .= "<tr><th>Start Date</th><th>Name</th><th>Machine</th><th>User</th>";
+      $out .= "<th>CPU </th><th>Memory (MB)</th><th>GPU</th><th>lifetime (h)</th><th>URL</th>";
+      $out .= "<th>PID's <form action=\"$config{http}://$config{server_name}/cgi-bin/qemu-web-desktop.pl\" method=post target=_blank><input type=hidden name=session_purge value=$user><input type=submit value=\"STOP ALL\"></input></form> </th></tr>\n";
+    }
+
+    foreach my $snapshot (glob("$dir/$service"."_*")) {
+      if (-d $snapshot) { # is a snapshot directory
+        my $snaphot_name = fileparse($snapshot); # just the session name
+        if (-e "$cfg/$snaphot_name.json") {
+          my $session_ref = session_load(\%config, "$cfg/$snaphot_name.json");
+          if ($session_ref) {
+            my %session = %{ $session_ref };
+            my $flag=0;
+            
+            if    ($iter == 1 and $user =~ $session{user}) { $flag = 1; }
+            elsif ($iter == 2 and $isadmin)                { $flag = 1; }
+            if ($session{pid} and $flag) {
+              my @pids = @{ $session{pid} };
+              $out .= "<tr>";
+              $out .= "<td>$session{date}</td>";
+              $out .= "<td>$session{name}</td>";
+              $out .= "<td>$session{machine}</td>";
+              $out .= "<td>$session{user}</td>";
+              $out .= "<td>$session{cpu}</td>";
+              $out .= "<td>$session{memory}</td>";
+              if ($session{gpu}) {
+                $out .= "<td>$session{gpu}</td>";
+              } else { $out .= "<td></td>"; }
+              my $lt = $session{snapshot_lifetime}/3600;
+              $out .= "<td>$lt</td>";
+              if ($session{oneshot}) {
+                $out .= "<td>one-shot";
+              } else {
+                $out .= "<td><a href='$session{url}' target=_blank><img alt=connect 
+          src=\"$config{http}://$session{server_name}/qemu-web-desktop/images/connect-button.jpeg\" 
+          title=Connect height=32></a>";
+                if ($session{token_vnc} && $config{service_port_multiple}) {
+                  $out .= " $session{token_vnc}";
+                }
               }
+              $out .= "</td>";
+              # create a STOP button for each
+              $out .= "<td>@pids <form action=\"$config{http}://$session{server_name}/cgi-bin/qemu-web-desktop.pl\" method=post target=_blank><input type=hidden name=session_stop value=\"$session{name}\"><input type=image alt=stop src=\"$config{http}://$session{server_name}/qemu-web-desktop/images/stop-button.png\" title=STOP height=32></form></td>\n";
             }
-            $out .= "</td>";
-            # create a STOP button for each
-            $out .= "<td>@pids <form action=\"/cgi-bin/qemu-web-desktop.pl\" method=post target=_blank><input type=hidden name=session_stop value=\"$session{name}\"><input type=submit value=STOP style=\"background-color:red\"></input></form></td>\n";
           }
         }
       }
-    }
-  }
-  $out .= "</table>";
-  
-  # display full list of sessions (when user is marked as 'admin')
-  if (not $isadmin) { return $out; }
-  
-  $out .= "<br><hr><br><h1>Active sessions [$config{session_nb}] (administration)</h1><table  border='1'>\n";
-  $out .= "<tr><th>Start Date</th><th>Name</th><th>Machine</th><th>User</th>";
-  $out .= "<th>CPU </th><th>Memory (MB)</th><th>GPU</th><th>lifetime (h)</th><th>URL</th>";
-  $out .= "<th>PID's <form action=\"/cgi-bin/qemu-web-desktop.pl\" method=post target=_blank><input type=hidden name=session_purge value=$user><input type=submit value=\"STOP ALL\"></input></form> </th></tr>\n";
-  foreach my $snapshot (glob("$dir/$service"."_*")) {
-    if (-d $snapshot) { # is a snapshot directory
-      my $snaphot_name = fileparse($snapshot); # just the session name
-      if (-e "$cfg/$snaphot_name.json") {
-        my $session_ref = session_load(\%config, "$cfg/$snaphot_name.json");
-        if ($session_ref) {
-          my %session = %{ $session_ref };
-          if ($session{pid}) {
-            my @pids = @{ $session{pid} };
-            $out .= "<tr>";
-            $out .= "<td>$session{date}</td>";
-            $out .= "<td>$session{name}</td>";
-            $out .= "<td>$session{machine}</td>";
-            $out .= "<td>$session{user}</td>";
-            $out .= "<td>$session{cpu}</td>";
-            $out .= "<td>$session{memory}</td>";
-            if ($session{gpu}) {
-              $out .= "<td>$session{gpu}</td>";
-            } else { $out .= "<td></td>"; }
-            my $lt = $session{snapshot_lifetime}/3600;
-            $out .= "<td>$lt</td>";
-            if ($session{oneshot}) {
-              $out .= "<td>one-shot";
-            } else {
-              $out .= "<td><a href='$session{url}' target=_blank><p style=\"color:green;\"><b>CONNECT</b></p></a>";
-              if ($session{token_vnc} && $config{service_port_multiple}) {
-                $out .= " $session{token_vnc}";
-              }
-            }
-            $out .= "</td>";
-            # create a STOP button for each
-            $out .= "<td>@pids <form action=\"/cgi-bin/qemu-web-desktop.pl\" method=post target=_blank><input type=hidden name=session_stop value=\"$session{name}\"><input type=submit value=STOP style=\"background-color:red\"></input></form></td>\n";
-          }
-        }
-      }
-    }
-  }
-  $out .= "</table>";
+    } # foreach snapshot
+    $out .= "</table>";
+    
+    # display full list of sessions (when user is marked as 'admin')
+    if (not $isadmin) { return $out; }
+  } # for 1..2
   
   return $out;
 }
@@ -1180,10 +1335,10 @@ sub session_email {
       $smtp->auth($config{email_from},$config{email_passwd}) || return;
     }
     $smtp->mail($config{email_from});
-    $smtp->recipient($session{user});
+    $smtp->recipient($session{user_email});
     $smtp->data();
     $smtp->datasend("From: $config{email_from}\n");
-    $smtp->datasend("To: $session{user}\n");
+    $smtp->datasend("To: $session{user_email}\n");
     # could add BCC to internal monitoring address $smtp->datasend("BCC: address\@example.com\n");
     $smtp->datasend("Subject: [Desktop] Remote $session{machine} connection information\n");
     $smtp->datasend("Content-Type: text/html; charset=\"UTF-8\" \n");
@@ -1197,6 +1352,47 @@ sub session_email {
 
 # ==============================================================================
 
+# service_check_user_max(\%config, $user)
+#   check resources used by user
+#   return: ($nb_session, $nb_cpu)
+sub service_check_user_max {
+
+  my $config_ref  = shift;
+  my $user        = shift;
+  
+  if (not $config_ref ) { return(0,0,0); }
+  my %config      = %{ $config_ref };
+  
+  my $nb_session = 0;
+  my $nb_cpu     = 0;
+  my $nb_mem     = 0;
+  
+  # get the list of active sessions for the given user
+  my $dir     = $config{dir_snapshots};
+  my $cfg     = $config{dir_cfg};
+  my $service = $config{service};
+  
+  foreach my $snapshot (glob("$dir/$service"."_*")) {
+    if (-d $snapshot) { # is a snapshot directory
+      my $snaphot_name = fileparse($snapshot); # just the session name
+      if (-e "$cfg/$snaphot_name.json") {
+        my $session_ref = session_load(\%config, "$cfg/$snaphot_name.json");
+        if ($session_ref) {
+          my %session = %{ $session_ref };
+          if ($session{pid} and $user =~ $session{user}) {
+            $nb_cpu     += $session{cpu};
+            $nb_session ++;
+            $nb_mem     += $session{memory};
+          }
+        }
+      }
+    }
+  }
+  
+  return ($nb_session, $nb_cpu, $nb_mem);
+  
+} # service_check_user_max
+
 # session_authenticate(\%config, \%session)
 #   check user credentials with SMTP, LDAP, IMAP and sendemail.
 #   return: ($out, $err)
@@ -1205,7 +1401,7 @@ sub session_authenticate {
   my $config_ref  = shift;
   my $session_ref = shift;
   
-  if (not $config_ref or not $session_ref) { return; }
+  if (not $config_ref or not $session_ref) { return ("",""); }
   my %config      = %{ $config_ref };
   my %session     = %{ $session_ref };
   
@@ -1222,6 +1418,7 @@ sub session_authenticate {
     if (index($authenticated, "SUCCESS") < 0 and $config{check_user_with_email} 
                            and Email::Valid->address($session{user})) {
       $authenticated = "EMAIL";
+      $session{user_email} = $session{user};
       
       $out .= "<li>[OK] An email will be sent to indicate the URL.</li>\n";
     }
@@ -1344,11 +1541,11 @@ sub session_check_ldap {
   my $config_ref  = shift;
   my $session_ref = shift;
 
-  if (not $config_ref or not $session_ref) { return; }
+  if (not $config_ref or not $session_ref) { return ""; }
   my %config      = %{ $config_ref };
   my %session     = %{ $session_ref };
 
-  if (not %config or not %session) { return; }
+  if (not %config or not %session) { return ""; }
   my $res    = "";
   my $filter = "";
 
@@ -1365,16 +1562,17 @@ sub session_check_ldap {
     
   # check if "user" was given as an email.
   if (Email::Valid->address($session{user})) {
+    $session{user_email} = $session{user};
     $filter = "mail=$session{user}";
   } else {
-    $filter = "cn=$session{user}";
+    $filter = "cn=$session{user}"; # may also be "uid=$session{user}"
   }
     
   # identify the DN
   my $mesg = $ldap->search(
     base => "dc=$config{ldap_domain}",
     filter => $filter,
-    attrs => ['dn','mail']);
+    attrs => ['dn','mail','cn','uid']);
   
   if (not $mesg or not $mesg->count) {
     $res = "FAILED: [LDAP] empty LDAP search.\n";
@@ -1383,10 +1581,13 @@ sub session_check_ldap {
     foreach my $entry ($mesg->all_entries) {
       my $dn    = $entry->dn();
       my $email = $entry->get_value('mail');
+      my $cn    = $entry->get_value('cn'); # may also be 'uid'
       my $bmesg = $ldap->bind($dn,password=>$session{password});
       if ( $bmesg and $bmesg->code() == 0 ) {
         if (not Email::Valid->address($session{user}) && $email) {
-          $session{user} = $email;
+          $session{user_email} = $email;
+        } elsif (Email::Valid->address($session{user})) {
+          $session{user} = $cn;
         }
         $res = "SUCCESS: [LDAP] $session{user} authenticated.";
       }
@@ -1411,6 +1612,8 @@ sub session_use_gpu {
   my %config      = %{ $config_ref };
   my $cfg     = $config{dir_cfg};
   
+  if ($pci =~ $config{gpu_blacklist}) { return 1; } # PCI is black-listed
+  
   # scan JSON files
   foreach my $json (glob("$cfg/*.json")) {
     
@@ -1418,7 +1621,7 @@ sub session_use_gpu {
     my $session_ref = session_load(\%config, $json);
     if ($session_ref) {
       my %session = %{ $session_ref };
-      if ($session{gpu} =~ $pci) { return 1; } # PCI is used
+      if ($session{gpu} =~ $pci) { return 1; }          # PCI is used
     }
   }
   return 0; # no session is using that PCI slot
