@@ -1,4 +1,10 @@
-# Configuration fon qemu-web-desktop
+# /etc/qemu-web-desktop/config.pl: Configuration for qemu-web-desktop
+#
+# Think about editing the /usr/share/qemu-web-desktop/html/desktop/index.html
+# to comment/un-comment sections (GPU, oneshot, user scripts, ...).
+#
+# When relevant, the equivalent OpenStack naming is indicated (as documented 
+# in https://docs.openstack.org/nova/latest/configuration/config.html)
 
 # WHERE THINGS ARE -------------------------------------------------------------
 
@@ -16,7 +22,7 @@ $config{dir_service}              = "/var/lib/$config{service}";
 # full path to machines (ISO,VM)
 $config{dir_machines}             = "$config{dir_service}/machines";
 
-# full path to snapshots and temporary files
+# full path to snapshots and temporary files OpenStack DEFAULT.instances_path
 $config{dir_snapshots}            = "$config{dir_service}/snapshots";
 
 # full path to snapshot config/lock files. Must NOT be accessible from http://
@@ -26,24 +32,13 @@ $config{dir_snapshots}            = "$config{dir_service}/snapshots";
 #   /etc/systemd/system/multi-user.target.wants/apache2.service
 #   PrivateTmp=true
 # which creates a '/tmp' in e.g. /tmp/systemd-private-*-apache2.service-*/
-$config{dir_cfg}                  = File::Spec->tmpdir(); 
+$config{dir_cfg}                  = File::Spec->tmpdir();  # OpenStack DEFAULT.tempdir
 
 # full path to noVNC, e.g. "/usr/share/novnc". Must contain "vnc.html"
 $config{dir_novnc}                = "/usr/share/novnc";
 # websockify command, e.g. "/usr/bin/websockify"
 $config{dir_websockify}           = "websockify";
 
-# search detached GPU (via vfio-pci). Only use video part, no audio.
-{
-  # can be a VGA adaptor or a 'controller'
-  my ($device_pci1, $device_model1, $device_name1) = pci_devices("lspci -nnk","vga","vfio");
-  my ($device_pci2, $device_model2, $device_name2) = pci_devices("lspci -nnk","controller","vfio");
-  
-  $config{gpu_model}             = ($device_model1, $device_model2);
-  $config{gpu_name}              = ($device_name1,  $device_name2);
-  $config{gpu_pci}               = ($device_pci1,   $device_pci2);
-}
-
 # MACHINE DEFAULT SETTINGS -----------------------------------------------------
 
 # max session life time in sec. 1 day is 86400 s. Highly recommended.
@@ -51,13 +46,13 @@ $config{dir_websockify}           = "websockify";
 $config{snapshot_lifetime}        = 86400*4; 
 
 # default nb of CPU per session.
-$config{snapshot_alloc_cpu}       = 1;
+$config{snapshot_alloc_cpu}       = 1; # OpenStack DEFAULT.initial_cpu_allocation_ratio ?
 
 # default nb of RAM per session (in GB).
-$config{snapshot_alloc_mem}       = 4;
+$config{snapshot_alloc_mem}       = 4; # OpenStack DEFAULT.initial_ram_allocation_ratio ?
 
 # default size of disk per session (in GB). Only for ISO machines.
-$config{snapshot_alloc_disk}      = 10.0;
+$config{snapshot_alloc_disk}      = 10.0; # OpenStack DEFAULT.initial_disk_allocation_ratio ?
 
 # default machine to run
 $config{machine}                  = 'slax.iso';
@@ -68,7 +63,7 @@ $config{qemu_exec}                = "qemu-system-x86_64";
 # QEMU video driver, can be "qxl" or "vmware"
 $config{qemu_video}               = "qxl"; 
 
-# Boot delay before displying the URL [sec]
+# Boot delay before displaying the URL [sec]
 $config{boot_delay}               = 5;
 
 # set a list of mounts to export into VMs.
@@ -84,22 +79,59 @@ $config{dir_mounts}               = [@mounts];
 $config{service_max_load}         = 0.8  ;
 
 # max number of active sessions. Deny service when above.
-$config{service_max_session_nb}   = 10;
+$config{service_max_session_nb}   = 10; # OpenStack config DEFAULT.max_concurrent_snapshots ?
+
+# max number of active sessions per user. Deny service when above.
+$config{service_max_session_nb_per_user}   = 3;
+
+# max number of active CPU fraction per user. Deny service when above.
+$config{service_max_cpu_fraction_nb_per_user}= 0.3;
+
+# max number of active memory fraction per user. Deny service when above.
+$config{service_max_mem_fraction_nb_per_user}= 0.3;
 
 # the base port for the VNC screen. 
 # the VNC port is chosen randomly up to service_port_vnc+service_max_session_nb
-$config{service_port_vnc}         = 5901;
+$config{service_port_vnc}         = 5901; # OpenStack config vmware.vnc_port
 
 # the port where the VNC screens are broadcasted (e.g. 6080 for websockify)
 # use 0 to select a random base port
-$config{service_port}             = 6080;
+$config{service_port}             = 6080; # OpenStack vnc.novncproxy_port
 $config{service_port_multiple}    = 0;  # when true, use one websockify per instance
-# 0: websockify uses only one port
-# 1: the websockify port is chosen randomly up to service_port+service_max_session_nb
+  # 0: websockify uses only one port (recommended, except when using 'oneshot' sessions)
+  # 1: the websockify port is chosen randomly up to service_port+service_max_session_nb
 
 # in case there is a proxy server to pass, specify it here
 $config{service_proxy}            = ""; # "http://195.221.0.35:8080/";
 
+# a list of other servers to be used when current one does not have enough resources
+# can be a server name, IP, including protocol (http, https). Indicate a 
+# comma-separated list. All of URL, server names and IP are allowed. 
+# e.g. 'https://grades-01.synchrotron-soleil.fr,https://re-grades-01.exp.synchrotron-soleil.fr';
+$config{fallback_servers}         = ''; 
+
+# the max length of user customization script (URL length or direct exec: commands)
+$config{service_max_script_length}= 65535; # e.g. 65535
+
+# Scripts to execute when any VM starts. Supports Linux and Windows guests.
+# see: https://libguestfs.org/virt-customize.1.html --commands-from-file
+# Can be specifyed as:
+#   "http://some/url"
+#   "/some/local/path/to/script"
+#   "exec: some commands" separated with EOL or ';'
+#   "virt-customize: commands" separated with EOL
+# The symbols `@USER@` `@SESSION_NAME@` and `@VM@` are replaced by the user 
+# name, the session ID, and the virtual machine name. 
+# They are executed at boot, in background, as root.
+# It is possible to define as many scripts as needed. Strings are separated by ','
+my @config_scripts=("");
+$config{config_script}            = [@config_scripts];
+
+# a list of GPU PCI addresses NOT to be used, as returned by lspci e.g. 
+# "06:05.0,00:1b.0". Items can be separated by spaces or commas.
+$config{gpu_blacklist}            = "";
+
+
 # USER AUTHENTICATION ----------------------------------------------------------
 
 # the name of the SMTP server, and optional port.
@@ -157,6 +189,7 @@ $config{check_user_with_ldap}     = 0;
 
 # Encryption (HTTPS). Set these to valid 'crt' and 'key' files with proper certificates
 # to allow secured encryption of websockify/novnc. The CRT and KEY should both be set.
+# OpenStack DEFAULT.key DEFAULT.cert
 $config{certificate_crt} = ''; # '/etc/apache2/certificate/apache-certificate.crt';
 $config{certificate_key} = ''; # '/etc/apache2/certificate/apache.key';
 
